/*
** $Id: calc.c, initiated June 20, 2008 $
** Calculus library
** See Copyright Notice in agena.h
*/


#include <stdlib.h>
#include <math.h>

#define calc_c
#define LUA_LIB

#include "../src/agena.h"

#include "../src/aauxlib.h"
#include "../src/agenalib.h"


#ifndef LUA_DOS
#define AGENA_CALCLIBNAME "calc"
LUALIB_API int (luaopen_calc) (lua_State *L);
#endif

static int calc_fsum (lua_State *L) {
  /* computes the sum of f(a), ..., f(b); 0.10.0 - April 20, 2008;
     12 % faster than an Agena implementation */
  lua_Number a, b, r;
  r = 0;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = agn_checknumber(L, 2);
  b = agn_checknumber(L, 3);
  while (a <= b) {
    lua_pushvalue(L, 1);
    lua_pushnumber(L, a);
    r += agn_ncall(L, 1, 1);
    a++;
  }
  lua_pushnumber(L, r);
  return 1;
}


static int calc_fseq (lua_State *L) {
  /* creates a sequence with values f(a), ..., f(b); June 30, 2008;
     33 % faster than an Agena implementation */
  lua_Number a, b, c;
  int i;
  c = 0;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = agn_checknumber(L, 2);
  b = agn_checknumber(L, 3);
  if (a > b)
    luaL_error(L, "in " LUA_QS ", start value greater than stop value.", "calc.fseq");
  agn_createseq(L, b-a+1);
  for (i=a; i<=b; i++) {
    lua_pushvalue(L, 1);
    lua_pushnumber(L, i);
    lua_seqsetinumber(L, -1, ++c, agn_ncall(L, 1, 1));
  }
  return 1;
}


static const luaL_Reg calclib[] = {
  {"fsum", calc_fsum},                     /* added on April 20, 2008 */
  {"fseq", calc_fseq},                     /* added on June 30, 2008 */
  {NULL, NULL}
};


/*
** Open calculus library
*/
LUALIB_API int luaopen_calc (lua_State *L) {
  luaL_register(L, AGENA_CALCLIBNAME, calclib);
  return 1;
}


