/*
** $Id: chelpers.h $
** C Helper routines
** See Copyright Notice in agena.h
** initiated Alexander Walz, July 20, 2007
*/

/* misc functions */

int tools_isPrime (unsigned int number);
double tools_intpow (double x, int n);
#ifndef LUA_ANSI
#include <complex.h>
complex double tools_centier (complex double x);
#endif

/* used in packages `ads` */

void tools_quicksort (long v[], int left, int right);

/* Dr. F.H.Toor's and my database C functions used in packages `base` and `phonetiqs` */

int my_open (const char *file);
int my_roopen (const char *file);
int my_create (const char *file) ;
int my_close (int hnd);
int my_seek (int hnd, long pos);
long my_lof (int hnd);
void my_read (int hnd, void *data, size_t size);
long my_readl (int hnd);
char my_readc (int hnd);
void my_write (int hnd, void *data, size_t size);
void my_writel (int hnd, long data);
void my_writec (int hnd, char data);
void my_lock (int hnd, long nbytes);
void my_unlock (int hnd, long nbytes);
void my_move (int hnd, long fpos, long tpos, long size);
void my_expand (int hnd, int mrc, int cnt, int count);
long my_fpos (int hnd);

size_t sec_read (int hnd, void *data, size_t size);
size_t sec_readl (int hnd, int *success);

/* get keystroke function*/
#ifdef __unix__
int getch(void);
#endif

const char *tools_cwd (void);

/* Endian functions */

char tools_endian (void);
void tools_swapdouble (double *n);
void tools_swaplong (long *n);

/* IEEE simulators */

int tools_isfinite (const double x);
int tools_isnan (const double x);
double tools_nan (void);


/* some constants */

#undef PI
#define PI (3.14159265358979323846)

#undef _E
#define _E (2.71828182845904523536)

#undef AGN_EPSILON
#define AGN_EPSILON  1.4901161193847656e-08  /* suited for doubles */

#undef RADIANS_PER_DEGREE
#define RADIANS_PER_DEGREE (PI/180.0)

#if defined(LUA_ANSI) || defined(_WIN32)
double csgn (double a, double b);
#endif

char *charreplace (char *s, char from, char to);


