/** $Id: lbaselib.c,v 1.191 15.03.2009 12:38:02 roberto Exp $
** Basic library
** See Copyright Notice in agena.h
*/

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>    /* added */

#ifdef _WIN32
#include <conio.h>   /* getch */
#include <io.h>      /* access in MingW */
#endif

#if defined(__unix__)
#include <unistd.h>  /* access in UNIX */
#endif

#if defined(__OS2__)
#include <conio.h>   /* getch */
#include <unistd.h>  /* access */
#endif

#include <math.h>    /* fabs, isnan, fmod, HUGE_VAL */

#define lbaselib_c
#define LUA_LIB

#include "agena.h"

#include "aauxlib.h"
#include "agenalib.h"
#include "loadlib.h"   /* for require subfunction fastload* */

#include "chelpers.h"  /* for Pi and Exp and charreplace */
#include "agncmpt.h"   /* for fmax */

/*
** If your system does not support `stdout', you can just remove this function.
** If you need, you can define your own `print' function, following this
** model but changing `fputs' to put the strings at a proper place
** (a console window or a log file, for instance).
*/

/* forward declaration */
void aux_printobject (lua_State *L, int i);

void aux_printtable (lua_State *L, int i) {
  int nops, getn, c, type, isutype;
  luaL_checktype(L, i, LUA_TTABLE);
  nops = agn_size(L, i);
  getn = lua_objlen(L, i);
  c = 0;
  isutype = agn_getutype(L, i);
  if (!isutype)
    fprintf(stdout, "[");
  else {
    fprintf(stdout, "%s(", lua_tostring(L, -1));
    agn_poptop(L);
  }
  lua_getglobal(L, "toString");
  lua_pushnil(L);
  while (lua_next(L, i-2) != 0) {
    c++;
    if (nops != getn) {
      /* print key */
      lua_pushvalue(L, -3);  /* toString */
      lua_pushvalue(L, -3);  /* key */
      lua_call(L, 1, 1);
      fprintf(stdout, "%s~", lua_tostring(L, -1));
      agn_poptop(L);  /* delete string */
    }
    /* print value */
    type = lua_type(L, -1);
    if (type != LUA_TNUMBER && type != LUA_TSTRING) {
      aux_printobject(L, -1);
    } else {
      lua_pushvalue(L, -3);  /* toString */
      lua_pushvalue(L, -2);  /* value */
      lua_call(L, 1, 1);
      fprintf(stdout, "%s", lua_tostring(L, -1));
      agn_poptop(L);  /* delete result (string) */
    }
    if (c < nops) fprintf(stdout, ", ");
    agn_poptop(L);  /* delete value */
  }
  if (isutype)
    fprintf(stdout, ")");
  else
    fprintf(stdout, "]");
  fflush(stdout);
  agn_poptop(L);  /* delete toString */
}


void aux_printset (lua_State *L, int i) {
  int nops, c, type, isnoutype;
  luaL_checktype(L, i, LUA_TSET);
  nops = agn_ssize(L, i);
  c = 0;
  if ((isnoutype = !agn_getutype(L, i)))
    fprintf(stdout, "{");
  else {
    fprintf(stdout, "%s(", lua_tostring(L, -1));
    agn_poptop(L);
  }
  lua_getglobal(L, "toString");
  lua_pushnil(L);
  while (lua_usnext(L, i-2) != 0) {
    c++;
    /* print item */
    lua_pushvalue(L, -3);  /* toString */
    lua_pushvalue(L, -3);  /* key */
    type = lua_type(L, -1);
    if (type != LUA_TNUMBER && type != LUA_TSTRING) {
      aux_printobject(L, -1);
      agn_poptop(L);
    } else {
      lua_call(L, 1, 1);
      fprintf(stdout, "%s", lua_tostring(L, -1));
    }
    if (c < nops) fprintf(stdout, ", ");
    lua_pop(L, 2);  /* delete string and value */
  }
  if (isnoutype)
    fprintf(stdout, "}");
  else
    fprintf(stdout, ")");
  agn_poptop(L);  /* delete toString */
  fflush(stdout);
}


void aux_printseq (lua_State *L, int i) {
  int j, type;
  size_t nops = agn_seqsize(L, i);
  if (!agn_getutype(L, i))
    fprintf(stdout, "seq(");
  else {
    fprintf(stdout, "%s(", lua_tostring(L, -1));
    agn_poptop(L);
  }
  for (j=0; j<nops; j++) {
    lua_getglobal(L, "toString");
    lua_seqgeti(L, i-1, j+1);
    type = lua_type(L, -1);
    if (type != LUA_TNUMBER && type != LUA_TSTRING) {
      aux_printobject(L, -1);
      agn_poptop(L);
    } else {
      lua_call(L, 1, 1);
      fprintf(stdout, "%s", lua_tostring(L, -1));
    }
    if (j < nops - 1)
      fprintf(stdout, ", ");
    agn_poptop(L);
  }
  fprintf(stdout, ")");
  fflush(stdout);
}


void aux_printpair (lua_State *L, int i) {
  int j, f, type;
  f = agn_getutype(L, i);  /* user-defined type set ? */
  if (f) {
    fprintf(stdout, "%s(", lua_tostring(L, -1));
    agn_poptop(L);
  }
  for (j=0; j<2; j++) {
    lua_getglobal(L, "toString");
    agn_pairgeti(L, i-1, j+1);
    type = lua_type(L, -1);
    if (type != LUA_TNUMBER && type != LUA_TSTRING) {
      aux_printobject(L, -1);
      agn_poptop(L);
      goto label;
    }
    lua_call(L, 1, 1);
    fprintf(stdout, "%s", lua_tostring(L, -1));
  label:
    if (j < 1) {
      if (f)
        fprintf(stdout, ", ");
      else
        fprintf(stdout, ":");
    }
    agn_poptop(L);
  }
  if (f) fprintf(stdout, ")");
  fflush(stdout);
}


void aux_printcomplex (lua_State *L, int i) {
  int isrealzero = 1, round = 0;
  lua_Number re, im, eps;
  lua_getglobal(L, "_EnvPrintZeroedCmplxVals");
  round = agn_istrue(L, -1);  /* round real and imaginary parts close to zero to zero on output */
  agn_poptop(L);
  lua_getglobal(L, "_EnvEps");
  if (lua_isnumber(L, -1))
    eps = agn_tonumber(L, -1);  /* round real and imaginary parts close to zero to zero on output */
  else
    eps = AGN_EPSILON;
  agn_poptop(L);
  agn_complexgetreal(L, i);
  re = lua_tonumber(L, -1);
  if (round && fabs(re) < eps) re = 0;
  if (re != 0) {
    fprintf(stdout, "%s", lua_tostring(L, -1));
    isrealzero = 0;
  }
  agn_poptop(L);
  agn_complexgetimag(L, i);
  im = lua_tonumber(L, -1);
  if (round && fabs(im) < eps) im = 0;
  if (im < 0) {
    if (im == -1)
      fprintf(stdout, "-I");
    else
      fprintf(stdout, "%s*I", lua_tostring(L, -1));
  } else if (im > 0) {
    if (im == 1)
      fprintf(stdout, "%sI", (isrealzero) ? "" : "+");
    else
      fprintf(stdout, "%s%s*I", (isrealzero) ? "" : "+", lua_tostring(L, -1));
  } else if (re == 0 && im == 0)
    fprintf(stdout, "0");
  else if ((im == 0) || (isnan(re) && isnan(im))) {}  /* do nothing */
  else
    fprintf(stdout, "PLEASE REPORT THIS BUG: real=%lf, imag=%lf", re, im);
  agn_poptop(L);
  fflush(stdout);
}

/* assumes that i is negative ! */
void aux_printnonstruct (lua_State *L, int i) {
  lua_getglobal(L, "toString");
  lua_pushvalue(L, i-1);
  lua_call(L, 1, 1);  /* this will also call the __string metamethod if one exists */
  fprintf(stdout, "%s", lua_tostring(L, -1));
  agn_poptop(L);  /* pop string */
  fflush(stdout);
}


int pagebreak (int couter, int more) {
  int key;
  if (fmod(couter, more) == 0) {
    fprintf(stderr, "more ...");
    key = getch();
    fprintf(stderr, "\b\b\b\b\b\b\b\b        \b\b\b\b\b\b\b\b");
    if (key == 27 || key == 81 || key == 113) return 1;
  }
  return 0;
}


void printnewline (lua_State *L) {
  lua_getglobal(L, "_EnvPrintNewLineAfterInput");
  if (!(lua_type(L, -1) == LUA_TBOOLEAN && agn_istrue(L, -1))) {
    fputs("\n", stdout);
  }
  agn_poptop(L);
}


void aux_Gprint (lua_State *L, int i, int flag, int more) {
  /* print contents of global _G, which is self-referencing, without getting a permanent loop
     flag = 1: print newlines between entries; flag = 0: print only commas
     more: number of lines printed before user prompt (only for flag = 1);
     0.12.0; July 01, 2008 */
  int c, nops;
  lua_pushvalue(L, i);  /* make `copy` of table */
  nops = agn_size(L, -1);
  lua_getglobal(L, "toString");
  lua_pushnil(L);
  c = 0;
  if (!flag) fprintf(stdout, "[");
  while (lua_next(L, -4) != 0) {
    c++;
    lua_pushvalue(L, -3);  /* push toString */
    lua_pushvalue(L, -3);  /* push key */
    lua_call(L, 1, 1);
    fprintf(stdout, "%s", lua_tostring(L, -1));
    agn_poptop(L);  /* delete result (string) */
    lua_pushvalue(L, -3);  /* push toString */
    lua_pushvalue(L, -2);  /* push value */
    lua_call(L, 1, 1);
    fprintf(stdout, " ~ %s", lua_tostring(L, -1));
    if (flag)
      fprintf(stdout, "\n");
    else
      fprintf(stdout, ", ");
    lua_pop(L, 2);  /* delete result (string) and value */
    if (flag && pagebreak(c, more) == 1) break;
  }
  if (!flag && c != 0) fprintf(stdout, "\b\b");
  lua_pop(L, 2);  /* toString and _G */
  if (!flag) fprintf(stdout, "]");
  printnewline(L);
}


void aux_printobject (lua_State *L, int i) {
  int pprinter;
  if ((pprinter = luaL_callmeta(L, i, "__tostring"))) { /* is there a metafield? */
    agn_poptop(L);  /* pop pretty printer */
  }
  switch (lua_type(L, i)) {
    case LUA_TTABLE:
      if (!pprinter) {
        /* avoid endless loop with _G */
        lua_pushvalue(L, i);  /* push table, patch 0.20.2 */
        lua_pushstring(L, "_G");  /* push _G key */
        lua_rawget(L, -2);  /* read table._G without invoking the __index metamethod */
        /* lua_getfield(L, i, "_G"); */
        if (!lua_isnil(L, -1)) {
          aux_Gprint(L, i-1, 0, 0);  /* push _G, leaves stack unchanged */
          agn_poptoptwo(L);     /* delete _G and table */
          break;
        }
        agn_poptoptwo(L);  /* drop _G and table */
        aux_printtable(L, i);
        break;
      }
    case LUA_TSET:
      if (!pprinter) {
        aux_printset(L, i);
        break;
      }
    case LUA_TSEQ:
      if (!pprinter) {
        aux_printseq(L, i);
        break;
      }
    case LUA_TPAIR:
      if (!pprinter) {
        aux_printpair(L, i);
        break;
      }
    case LUA_TCOMPLEX:
      if (!pprinter) {
        aux_printcomplex(L, i);
        break;
      }
    default:
      aux_printnonstruct(L, i);
  }
}


/* this macro checks whether there is a user defined printing function for the requested object
   and if so calls this function; otherwise the macro is `skipped`. */

#define auxuserprint(L, fn) { \
  lua_getfield(L, -1, fn); \
  if (lua_isfunction(L, -1)) { \
    lua_pushvalue(L, -3); \
    lua_call(L, 1, 0); \
    lua_pop(L, 2); \
    continue; \
  } else \
    agn_poptop(L); \
  break; \
}

static int luaB_print (lua_State *L) {
  int n = lua_gettop(L);  /* number of arguments */
  int i, flag, couter, more, pprinter, longtable;
  lua_getglobal(L, "_EnvMore");  /* number of lines to be printed until user interaction */
  more = lua_tonumber(L, -1);  /* if not a number, more is assigned HUGE_VAL */
  agn_poptop(L);
  lua_getglobal(L, "_EnvLongTable");  /* print table in long format, i.e. each key~value pair at a separate line */
  longtable = lua_toboolean(L, -1);
  agn_poptop(L);
  lua_getglobal(L, "_EnvPrint");
  if (lua_istable(L, -1)) {
    lua_getfield(L, -1, "Table");
    if (lua_isfunction(L, -1)) longtable = 0;  /* if there is a user-defined printing function, use it */
    agn_poptop(L);
  }
  agn_poptop(L);
  if ((pprinter = luaL_callmeta(L, 1, "__tostring"))) { /* is there a metafield? */
    agn_poptop(L);  /* pop pretty printer */
  }
  couter = 0;
  if (n == 1 && !pprinter && lua_type(L, 1) == LUA_TTABLE && longtable) {
    lua_pushvalue(L, 1);  /* push table, patch 0.20.2 */
    lua_pushstring(L, "_G");  /* push _G key */
    lua_rawget(L, -2);    /* do not invoke __index metamethod */
    /* lua_getfield(L, 1, "_G"); */
    if (!lua_isnil(L, -1)) {  /* avoid endless printing loop with _G (and thus also _origG) */
      aux_Gprint(L, -2, 1, more);
      agn_poptoptwo(L);  /* remove _G and table */
    }
    else {
      agn_poptoptwo(L);  /* remove nil and table */
      flag = 1;
      couter++;
      lua_pushnil(L);
      while (lua_next(L, 1) != 0) {
        flag = 0;
        aux_printobject(L, -2);  /* print key */
        fputs(" ~ ", stdout);
        aux_printobject(L, -1);  /* print value */
        fputs("\n", stdout);
        agn_poptop(L);  /* pop value */
        couter++;
        if (pagebreak(couter, more) == 1) break;
      }
      if (flag) fputs("[]\n", stdout);  /* empty table */
    }
    return 0;
  }
  for (i=1; i<=n; i++) {
    if (i>1) fputs("\t", stdout);
    lua_pushvalue(L, i);
    if (!pprinter) {  /* no __string metamethod ? */
      lua_getglobal(L, "_EnvPrint");
      if (lua_type(L, -1) == LUA_TTABLE) { /* are there user-defined formatting functions ? */
        switch (lua_type(L, -2)) {
          case LUA_TTABLE:
            auxuserprint(L, "Table");
          case LUA_TSET:
            auxuserprint(L, "Set");
          case LUA_TSEQ:
            auxuserprint(L, "Sequence");
          case LUA_TPAIR:
            auxuserprint(L, "Pair");
          case LUA_TCOMPLEX:
            auxuserprint(L, "Complex");
          default: break;
        }
      }
      agn_poptop(L);  /* remove _EnvPrint */
    }
    /* if no user-defined formatting function for the given structure could be found or if there
       is a `__string` metamethod for the object, then the next line will be executed. */
    aux_printobject(L, -1);  /* aux_printobject can only be called with a negative index */
    agn_poptop(L);
  }
  fputs("\n", stdout);
  return 0;
}


/* 0.12.0, July 01, 2008 */
static int luaB_dottyprint (lua_State *L) {
  /* save last value printed to `ans` */
  if (!lua_isnil(L, 1)) {
    lua_pushvalue(L, 1);
    lua_setglobal(L, "ans");  /* pops the value from the stack */
  }
  /* now print values (last value must be stored to ans because the pcall removes the values) */
  lua_getglobal(L, "print");
  lua_insert(L, 1);
  if (lua_pcall(L, lua_gettop(L)-1, 0, 0) != 0)
    luaL_error(L, "error calling " LUA_QS, "print");
  /* print leaves nothing on the stack, so do not pop anything */
  return 0;
}


static int luaB_mprint (lua_State *L) {
  int n = lua_gettop(L);  /* number of arguments */
  int i; int j;
  char space[2];  /* changed 0.5.3 */
  lua_getglobal(L, "toString");
  strcpy(space, " \0");
  for (i=1; i<=n; i++) {
    const char *s;
    lua_pushvalue(L, -1);  /* function to be called */
    lua_pushvalue(L, i);   /* value to print */
    lua_call(L, 1, 1);
    s = lua_tostring(L, -1);  /* get result */
    if (s == NULL)
      return luaL_error(L, LUA_QL("toString") " must return a string to "
                           LUA_QL("mprint"));
    if (n == 1) for (j=0; j<(80-strlen(s))/2; j++) fputs(space, stdout);
    if (i>1) fputs("\t", stdout);
    fputs(s, stdout);
    agn_poptop(L);  /* pop result */
  }
  printnewline(L);
  return 0;
}


static int luaB_toNumber (lua_State *L) {
  int base = luaL_optint(L, 2, 10);
  if (base == 10) {  /* standard conversion */
    lua_Number n;
    int exception;
    luaL_checkany(L, 1);
    if (lua_isstring(L, 1) || lua_isnumber(L, 1)) {  /* changed 0.12.2 */
      n = agn_tonumberx(L, 1, &exception);
      if (exception)  /* conversion failed ? */
        lua_pushvalue(L, 1);
      else
        lua_pushnumber(L, n);
      return 1;
    }
  }
  else {
    const char *s1 = luaL_checkstring(L, 1);
    char *s2;
    unsigned long n;
    luaL_argcheck(L, 2 <= base && base <= 36, 2, "base out of range");
    n = strtoul(s1, &s2, base);
    if (s1 != s2) {  /* at least one valid digit? */
      while (isspace((unsigned char)(*s2))) s2++;  /* skip trailing spaces */
      if (*s2 == '\0') {  /* no invalid trailing characters? */
        lua_pushnumber(L, (lua_Number)n);
        return 1;
      }
    }
  }
  lua_pushfail(L);  /* else not a number */
  return 1;
}


static int luaB_error (lua_State *L) {
  int level = luaL_optint(L, 2, 1);
  lua_settop(L, 1);
  if (lua_isstring(L, 1) && level > 0) {  /* add extra information? */
    luaL_where(L, level);
    lua_pushvalue(L, 1);
    lua_concat(L, 2);
  }
  return lua_error(L);
}


static int luaB_getmetatable (lua_State *L) {
  luaL_checkany(L, 1);
  if (!lua_getmetatable(L, 1)) {
    lua_pushnil(L);
    return 1;  /* no metatable */
  }
  luaL_getmetafield(L, 1, "__metatable");
  return 1;  /* returns either __metatable field (if present) or metatable */
}


static int luaB_setmetatable (lua_State *L) {
  int s = lua_type(L, 1);
  int t = lua_type(L, 2);
  luaL_argcheck(L, s == LUA_TTABLE || s == LUA_TSEQ || s == LUA_TSET || s == LUA_TPAIR , 1,
                    "table, set, sequence, or pair expected");
  luaL_argcheck(L, t == LUA_TNIL || t == LUA_TTABLE, 2,
                    "null or table expected");
  if (luaL_getmetafield(L, 1, "__metatable"))
    luaL_error(L, "to " LUA_QS ", cannot change a protected metatable", "setmetatable");
  lua_settop(L, 2);
  lua_setmetatable(L, 1);
  return 1;
}


static void getfunc (lua_State *L, int opt) {  /* Lua 5.1.2 patch */
  if (lua_isfunction(L, 1)) lua_pushvalue(L, 1);
  else {
    lua_Debug ar;
    int level = opt ? luaL_optint(L, 1, 1) : luaL_checkint(L, 1);  /* Lua 5.1.2 patch */
    luaL_argcheck(L, level >= 0, 1, "level must be non-negative");
    if (lua_getstack(L, level, &ar) == 0)
      luaL_argerror(L, 1, "invalid level");
    lua_getinfo(L, "f", &ar);
    if (lua_isnil(L, -1))
      luaL_error(L, "no function environment for tail call at level %d",
                    level);
  }
}


static int luaB_getfenv (lua_State *L) {
  getfunc(L, 1); /* Lua 5.1.2 patch */
  if (lua_iscfunction(L, -1))  /* is a C function? */
    lua_pushvalue(L, LUA_GLOBALSINDEX);  /* return the thread's global env. */
  else
    lua_getfenv(L, -1);
  return 1;
}


static int luaB_setfenv (lua_State *L) {
  luaL_checktype(L, 2, LUA_TTABLE);
  getfunc(L, 1);  /* Lua 5.1.2 patch */
  lua_pushvalue(L, 2);
  if (lua_isnumber(L, 1) && agn_tonumber(L, 1) == 0) {
    /* change environment of current thread */
    lua_pushthread(L);
    lua_insert(L, -2);
    lua_setfenv(L, -2);
    return 0;
  }
  else if (lua_iscfunction(L, -2) || lua_setfenv(L, -2) == 0)
    luaL_error(L,
          LUA_QL("setfenv") " cannot change environment of given object");
  return 1;
}


static int luaB_rawequal (lua_State *L) {
  luaL_checkany(L, 1);
  luaL_checkany(L, 2);
  lua_pushboolean(L, lua_rawequal(L, 1, 2));
  return 1;
}


static int luaB_rawget (lua_State *L) {
  int s = lua_type(L, 1);
  luaL_checkany(L, 2);
  lua_settop(L, 2);
  switch (s) {
    case LUA_TTABLE:
      lua_rawget(L, 1);
      break;
    case LUA_TSET:
      lua_srawget(L, 1);
      break;
    case LUA_TSEQ:
      lua_seqrawget(L, 1);
      break;
    case LUA_TPAIR:
      agn_pairrawget(L, 1);
      break;
    default:
      luaL_error(L, "in " LUA_QS ", table, set, sequence, or pair expected.", "rawget");
  }
  return 1;
}


/* the switch/case version is a little bit slower */
static int luaB_rawset (lua_State *L) {
  int s = lua_type(L, 1);
  luaL_checkany(L, 2);
  /* luaL_checkany(L, narg): Checks whether the function has an argument of
     any type (including nil) at position narg. */
  if (s == LUA_TTABLE) {
    if (lua_gettop(L) == 2) {  /* 0.21.0, index missing ? */
      lua_pushinteger(L, agn_size(L, 1) + 1);  /* determine next free index */
      lua_insert(L, 2);  /* move it into arg position #2 */
    } else
      luaL_checkany(L, 3);
    lua_settop(L, 3);
    lua_rawset(L, 1);
  }
  else if (s == LUA_TSET) {
    lua_settop(L, 2);
    lua_srawset(L, 1);
  }
  else if (s == LUA_TSEQ) {
    if (lua_gettop(L) == 2) {  /* 0.21.0, index missing ? */
      lua_pushinteger(L, agn_seqsize(L, 1) + 1);  /* determine next free index */
      lua_insert(L, 2);  /* move it into arg position #2 */
    } else
      luaL_checkany(L, 3);
    lua_settop(L, 3);
    lua_seqrawset(L, 1);
  }
  else if (s == LUA_TPAIR) {
    luaL_checkany(L, 3);
    lua_settop(L, 3);
    agn_pairrawset(L, 1);
  }
  else
    luaL_error(L, "in " LUA_QS ", table, set, sequence, or pair expected.", "rawset");
  return 1;
}


static int luaB_gcinfo (lua_State *L) {
  lua_pushinteger(L, lua_getgccount(L));
  return 1;
}


static int luaB_collectgarbage (lua_State *L) {
  static const char *const opts[] = {"stop", "restart", "collect",
    "count", "step", "setpause", "setstepmul", NULL};
  static const int optsnum[] = {LUA_GCSTOP, LUA_GCRESTART, LUA_GCCOLLECT,
    LUA_GCCOUNT, LUA_GCSTEP, LUA_GCSETPAUSE, LUA_GCSETSTEPMUL};
  int o = luaL_checkoption(L, 1, "collect", opts);
  int ex = luaL_optint(L, 2, 0);
  int res = lua_gc(L, optsnum[o], ex);
  switch (optsnum[o]) {
    case LUA_GCCOUNT: {
      int b = lua_gc(L, LUA_GCCOUNTB, 0);
      lua_pushnumber(L, res + ((lua_Number)b/1024));
      return 1;
    }
    case LUA_GCSTEP: {
      lua_pushboolean(L, res);
      return 1;
    }
    default: {
      lua_pushnumber(L, res);
      return 1;
    }
  }
}


static int luaB_next (lua_State *L) {  /* 0.10.0: extended to handle sets, as well */
  switch (lua_type(L, 1)) {
    case LUA_TTABLE: {
      lua_settop(L, 2);  /* create a 2nd argument if there isn't one */
      if (lua_next(L, 1))
        return 2;
      else {
        lua_pushnil(L);
        return 1;
      }
    }
    case LUA_TSET: {
      lua_settop(L, 2);  /* create a 2nd argument if there isn't one */
      if (lua_usnext(L, 1))
        return 2;
      else {
        lua_pushnil(L);
        return 1;
      }
    }
    case LUA_TSEQ: {
      lua_settop(L, 2);  /* create a 2nd argument if there isn't one */
      if (lua_seqnext(L, 1))
        return 2;
      else {
        lua_pushnil(L);
        return 1;
      }
    }
    case LUA_TSTRING: {
      lua_settop(L, 2);  /* create a 2nd argument if there isn't one */
      if (lua_strnext(L, 1))
        return 2;
      else {
        lua_pushnil(L);
        return 1;
      }
    }
    default: {
      luaL_error(L, "in " LUA_QS ", table, set, sequence, or string expected.", "next");
    }
  }
  return 1;
}


static int load_aux (lua_State *L, int status) {
  if (status == 0)  /* OK? */
    return 1;
  else {
    lua_pushnil(L);
    lua_insert(L, -2);  /* put before error message */
    return 2;  /* return nil plus error message */
  }
}


static int luaB_loadstring (lua_State *L) {
  size_t l;
  const char *s = luaL_checklstring(L, 1, &l);
  const char *chunkname = luaL_optstring(L, 2, s);
  return load_aux(L, luaL_loadbuffer(L, s, l, chunkname));
}


static int luaB_loadfile (lua_State *L) {
  const char *fname = luaL_optstring(L, 1, NULL);
  return load_aux(L, luaL_loadfile(L, fname));
}


/*
** Reader for generic `load' function: `lua_load' uses the
** stack for internal stuff, so the reader cannot change the
** stack top. Instead, it keeps its resulting string in a
** reserved slot inside the stack.
*/
static const char *generic_reader (lua_State *L, void *ud, size_t *size) {
  (void)ud;  /* to avoid warnings */
  luaL_checkstack(L, 2, "too many nested functions");
  lua_pushvalue(L, 1);  /* get function */
  lua_call(L, 0, 1);  /* call it */
  if (lua_isnil(L, -1)) {
    *size = 0;
    return NULL;
  }
  else if (lua_isstring(L, -1)) {
    lua_replace(L, 3);  /* save string in a reserved stack slot */
    return lua_tolstring(L, 3, size);
  }
  else luaL_error(L, "reader function must return a string");
  return NULL;  /* to avoid warnings */
}


static int luaB_load (lua_State *L) {
  int status;
  const char *cname = luaL_optstring(L, 2, "=(load)");
  luaL_checktype(L, 1, LUA_TFUNCTION);
  lua_settop(L, 3);  /* function, eventual name, plus one reserved slot */
  status = lua_load(L, generic_reader, NULL, cname);
  return load_aux(L, status);
}


static int luaB_run (lua_State *L) {
  const char *fname = luaL_optstring(L, 1, NULL);
  int n = lua_gettop(L);
  if (luaL_loadfile(L, fname) != 0) lua_error(L);
  lua_call(L, 0, LUA_MULTRET);
  return lua_gettop(L) - n;
}


static int luaB_assume (lua_State *L) {
  luaL_checkany(L, 1);
  if (!lua_toboolean(L, 1))
    return luaL_error(L, "%s", luaL_optstring(L, 2, "assumption failed"));
  return lua_gettop(L);
}


static int luaB_unpack (lua_State *L) {
  int i, e, n, o;
  o = lua_type(L, 1);
  luaL_argcheck(L, o == LUA_TTABLE || o == LUA_TSEQ, 1,
                    "table or sequence expected");
  i = luaL_optint(L, 2, 1);
  e = luaL_opt(L, luaL_checkint, 3, (o == LUA_TTABLE) ? luaL_getn(L, 1) : agn_seqsize(L, 1));
    /* 0.14.0, do not use agn_nops for it also counts hash values */
  if (i > e) return 0;  /* empty range; Lua 5.1.3 patch 4 */
  n = e - i + 1;  /* number of elements */
  if (n <= 0 || !lua_checkstack(L, n))  /* n <= 0 means arith. overflow */
    return luaL_error(L, "too many results to unpack");
  switch (o) {
    case LUA_TTABLE:
      lua_rawgeti(L, 1, i);  /* push arg[i] (avoiding overflow problems) */
      while (i++ < e)  /* push arg[i + 1...e] */
        lua_rawgeti(L, 1, i);
      break;
    case LUA_TSEQ:
      lua_seqgeti(L, 1, i);  /* push arg[i] (avoiding overflow problems) */
      while (i++ < e)  /* push arg[i + 1...e] */
        lua_seqgeti(L, 1, i);
      break;
    default:
      lua_assert(0); break;  /* shall not happen */
  }
  return n;
}


static int luaB_pcall (lua_State *L) {
  int status;
  luaL_checkany(L, 1);
  status = lua_pcall(L, lua_gettop(L) - 1, LUA_MULTRET, 0);
  lua_pushboolean(L, (status == 0));
  lua_insert(L, 1);
  return lua_gettop(L);  /* return status + all results */
}


static int luaB_xpcall (lua_State *L) {
  int status;
  luaL_checkany(L, 2);
  lua_settop(L, 2);
  lua_insert(L, 1);  /* put error function under function to be called */
  status = lua_pcall(L, 0, LUA_MULTRET, 1);
  lua_pushboolean(L, (status == 0));
  lua_replace(L, 1);
  return lua_gettop(L);  /* return status + all results */
}


static int luaB_toString (lua_State *L) {
  luaL_checkany(L, 1);
  if (luaL_callmeta(L, 1, "__tostring"))  /* is there a metafield? */
    return 1;  /* use its value */
  switch (lua_type(L, 1)) {
    case LUA_TNUMBER:
      lua_pushstring(L, lua_tostring(L, 1));
      break;
    case LUA_TSTRING:
      lua_pushvalue(L, 1);
      break;
    case LUA_TBOOLEAN:
      if (lua_toboolean(L, 1) == LUA_TFAIL)
        lua_pushstring(L, "fail");
      else
        lua_pushstring(L, (lua_toboolean(L, 1) ? "true" : "false"));
      break;
    case LUA_TNIL:
      lua_pushliteral(L, "null");  /* changed */
      break;
    case LUA_TCOMPLEX: {  /* 0.14.0 */
#ifndef LUA_ANSI
      agn_Complex z = agn_tocomplex(L, 1);
      lua_pushnumber(L, creal(z));
      lua_pushstring(L, lua_tostring(L, -1));
      lua_remove(L, -2);  /* delete creal(z) */
      lua_pushnumber(L, cimag(z));
      lua_pushstring(L, lua_tostring(L, -1));
      lua_remove(L, -2);  /* delete cimag(z) */
#else
      lua_Number a = agn_complexreal(L, 1), b = agn_compleximag(L, 1);
      lua_pushnumber(L, a);
      lua_pushstring(L, lua_tostring(L, -1));
      lua_remove(L, -2);  /* delete a */
      lua_pushnumber(L, b);
      lua_pushstring(L, lua_tostring(L, -1));
      lua_remove(L, -2);  /* delete a */
#endif
      return 2;
    }
    case LUA_TPAIR: {   /* 0.14.0 */
      lua_pushnumber(L, 1);  /* get left-hand side */
      agn_pairrawget(L, 1);  /* and put it onto the stack top, deleting 1 */
      lua_pushstring(L, lua_tostring(L, -1));
      lua_remove(L, -2);     /* remove left-hand side */
      lua_pushnumber(L, 2);  /* get right-hand side */
      agn_pairrawget(L, 1);  /* and put it onto the stack top, deleting 2 */
      lua_pushstring(L, lua_tostring(L, -1));
      lua_remove(L, -2);     /* remove right-hand side */
      return 2;
    }
    default:
      lua_pushfstring(L, "%s(%p)", luaL_typename(L, 1), lua_topointer(L, 1));
      break;
  }
  return 1;
}


static int luaB_newproxy (lua_State *L) {
  lua_settop(L, 1);
  lua_newuserdata(L, 0);  /* create proxy */
  if (lua_toboolean(L, 1) == 0)
    return 1;  /* no metatable */
  else if (lua_isboolean(L, 1)) {
    lua_newtable(L);  /* create a new metatable `m' ... */
    lua_pushvalue(L, -1);  /* ... and mark `m' as a valid metatable */
    lua_pushboolean(L, 1);
    lua_rawset(L, lua_upvalueindex(1));  /* weaktable[m] = true */
  }
  else {
    int validproxy = 0;  /* to check if weaktable[metatable(u)] == true */
    if (lua_getmetatable(L, 1)) {
      lua_rawget(L, lua_upvalueindex(1));
      validproxy = lua_toboolean(L, -1);
      agn_poptop(L);  /* remove value */
    }
    luaL_argcheck(L, validproxy, 1, "boolean or proxy expected");
    lua_getmetatable(L, 1);  /* metatable is valid; get it */
  }
  lua_setmetatable(L, 2);
  return 1;
}

/* ADDITIONS */


static int agn_time (lua_State *L) {
  lua_pushnumber(L, ((lua_Number)clock())/(lua_Number)CLOCKS_PER_SEC);
  return 1;
}


/* December 10, 2006; revised December 24, 2007; extended June 21, 2008, patched December 23, 2008
   Since values popped may be gc'ed, the strings put on the stack are only popped when not needed
   any longer. Do not put an agn_poptop() statement right after the str = lua_tostring(L, n)
   assignment !!! */
static int agn_readlib (lua_State *L) {
  const char *fn, *path;
  int nargs, i, *results;
  nargs = lua_gettop(L);
  results = malloc(nargs*sizeof(char));
  luaL_checkstack(L, nargs, "too many arguments");
  lua_getfield(L, LUA_GLOBALSINDEX, "_EnvAgenaPath");
  if (!lua_isstring(L, -1)) {  /* 0.20.2 */
    luaL_error(L, "Error, " LUA_QS
      " is not a string, must be a path to the main Agena folder", "_EnvAgenaPath");
  }
  path = luaL_checkstring(L, -1);
  if (path == NULL) {
    fprintf(stderr, "Warning in readlib: _EnvAgenaPath not set, cannot load package.\n");
    fflush(stderr);
    agn_poptop(L);  /* drop path (i.e. null) */
    lua_pushfail(L);
    return 1;
  }
  for (i=1; i <= nargs; i++) {
    const char *s;
    int result, clibsuccess;
    s = luaL_checkstring(L, i);
    clibsuccess = 0;
    result = luaL_loadfile(L, s);
    /* first try to load the _file_ `s` */
    if (result != LUA_ERRFILE) {  /* file found, run it */
      if (result != 0) lua_error(L); /* but there is a syntax error in it */
      if (lua_pcall(L, 0, LUA_MULTRET, 0) != 0) {  /* the loaded file must be run to get procs assigned */
        lua_error(L);
      } else
        goto res;
    } else
      agn_poptop(L);
    /* try finding and reading CLIB (DLLs, SOs) */
    lua_pushstring(L, path); lua_pushstring(L, "/lib/");
    lua_pushstring(L, s); lua_pushstring(L, AGN_CLIB);
    lua_concat(L, 4);
    fn = lua_tostring(L, -1);
    fastloader_C(L, fn, s, &clibsuccess);
    agn_poptop(L);  /* pop package table or fail */
    agn_poptop(L);  /* drop filename */
    /* load agn text file libraries if present */
    /* try to find file in /<packagename> folder */
    lua_pushstring(L, path); lua_pushstring(L, "/"); lua_pushstring(L, s);
    lua_pushstring(L, "/"); lua_pushstring(L, s); lua_pushstring(L, ".agn");
    lua_concat(L, 6);
    fn = lua_tostring(L, -1);
    result = luaL_loadfile(L, fn);
    if (result != LUA_ERRFILE) {  /* file found, run it */
      if (result != 0) lua_error(L); /* but there is a syntax error in it */
      if (lua_pcall(L, 0, LUA_MULTRET, 0) != 0) {  /* the loaded file must be run to get procs assigned */
        lua_error(L);
      } else
        goto res;
    } else
      agn_poptop(L);
    agn_poptop(L);  /* drop filename */
    /* try to find file in /lib folder */
    lua_pushstring(L, path); lua_pushstring(L, "/lib/");
    lua_pushstring(L, s); lua_pushstring(L, ".agn");
    lua_concat(L, 4);
    fn = lua_tostring(L, -1);
    result = luaL_loadfile(L, fn);
    if (result != LUA_ERRFILE) {  /* file found, run it */
      if (result != 0) lua_error(L);  /* but there is a syntax error in it */
      if (lua_pcall(L, 0, LUA_MULTRET, 0) != 0) {  /* the loaded file must be run to get procs assigned */
        lua_error(L);
      } else
        goto res;
    } else
      agn_poptop(L);
    agn_poptop(L);  /* drop filename */
    if (result && !clibsuccess) {
      fprintf(stderr, "Error, package or file " LUA_QS " not found.\n", s);
      fflush(stderr);
    }
  res:  /* don't try lua_pushboolean(result) here because fastloaderC pushes values on stack that apparently
     must not be deleted; instead collect result of readlib attempt in an array */
    results[i-1] = !result || clibsuccess;
  }
  agn_poptop(L);  /* drop path */
  for (i=0; i < nargs; i++) {
    if (results[i])
      lua_pushtrue(L);
    else
      lua_pushfail(L);
  }
  free(results);
  return nargs;
}


static int agn_loadClib (lua_State *L) {
  const char *fn, *s, *dirpath;
  int success;
  s = luaL_checkstring(L, 1);  /* name of package */
  dirpath = luaL_checkstring(L, 2);  /* path to the package (no filename, no trailing slash !) */
  /* try finding and reading CLIB (DLLs, SOs) */
  lua_pushstring(L, dirpath); lua_pushstring(L, "/");
  lua_pushstring(L, s); lua_pushstring(L, AGN_CLIB);
  lua_concat(L, 4);
  fn = lua_tostring(L, -1);
  fastloader_C(L, fn, s, &success);
  lua_remove(L, -2);  /* remove filename fn */
  if (!success) agn_poptop(L);  /* remove fail returned by fastloader_C */
  lua_pushboolean(L, success);
  return 1;
}


int agn_restart (lua_State *L) { /* changed 0.4.0 */
  const char *path;
  /* first delete all global vars except _origG and _EnvAgenaPath */
  lua_getfield(L, LUA_GLOBALSINDEX, "_G");  /* get _G on stack */
  lua_pushnil(L);
  while (lua_next(L, -2) != 0) {
    lua_pushnil(L);
    if (strcmp(lua_tostring(L, -3), "_origG") != 0 &&
        strcmp(lua_tostring(L, -3), "_EnvAgenaPath") != 0) {
      lua_setfield(L, LUA_GLOBALSINDEX, lua_tostring(L, -3));
      agn_poptop(L);
      } /* delete key in Agena environment */
    else {
      agn_poptoptwo(L);
    }
  }
  agn_poptop(L); /* delete _G from stack */
  /* delete entries of loaded Lua modules, 0.12.0 */
  lua_pushnil(L);
  lua_setfield(L, LUA_REGISTRYINDEX, "_LOADED");
  /* get _oldG on stack */
  lua_getfield(L, LUA_GLOBALSINDEX, "_origG");
  if (!lua_istable(L, -1)) {
    fprintf(stderr, "Warning, `_origG` is not a table, no restart possible.\n");
    fflush(stderr);
    agn_poptop(L);
    lua_pushfail(L);
    return 1;
  }
  /* set all values in _origG into environment */
  lua_pushnil(L);
  while (lua_next(L, -2) != 0) {
    lua_pushvalue(L, -2); /* key */
    lua_pushvalue(L, -2); /* value */
    lua_setfield(L, LUA_GLOBALSINDEX, lua_tostring(L, -2)); /* set key in Agena environment */
    agn_poptoptwo(L);
  }
  /* make Agena assign _G properly in the restarted environment. */
  lua_pushvalue(L, LUA_GLOBALSINDEX);
  lua_setglobal(L, "_G");
  /* get _EnvAgenaPath */
  agenaL_setAgenaPath(L, 0);
  lua_getfield(L, LUA_GLOBALSINDEX, "_EnvAgenaPath");
  path = lua_tostring(L, -1);
  if (path == NULL) {
    fprintf(stderr, "Warning, _EnvAgenaPath could not be determined, initialisation unsuccessful.\n");
    fflush(stderr);
  }
  else {
    int result;
    const char *fn;
    /* read library.agn */
    lua_pushstring(L, path); lua_pushstring(L, "/lib/library.agn");
    lua_concat(L, 2);
    fn = lua_tostring(L, -1);
    if (access(fn, 00) != -1) {  /* no error if ini file was not found */
      result = luaL_dofile(L, fn);  /* dofile leaves nothing on the stack */
      if (result != 0)  /* syntax error */
        lua_error(L);
    }
    agn_poptop(L);  /* pop fn */
    /* read agena.ini */
    lua_pushstring(L, path); lua_pushstring(L, "/lib/agena.ini");
    lua_concat(L, 2);
    fn = lua_tostring(L, -1);
    if (access(fn, 00) != -1) {  /* no error if ini file was not found */
      result = luaL_dofile(L, fn);  /* dofile leaves nothing on the stack */
      if (result != 0)  /* syntax error */
        lua_error(L);
    }
    agn_poptop(L);  /* pop fn */
  }
  agn_poptop(L);  /* drop path */
  lua_gc(L, LUA_GCCOLLECT, 0);
  lua_pushtrue(L);
  return 1;
}


/* 0.10.0, this one may be deprecated in the future */
static int agnB_maptoset (lua_State *L) {
  int i, j, n, nargs;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  nargs = lua_gettop(L);
  luaL_checkstack(L, nargs, "too many arguments");
  switch (lua_type(L, 2)) {
    case LUA_TTABLE: {
      n = luaL_getn(L, 2);
      if (n != 0) {  /* simple table */
        agn_createset(L, n);
        for (i=1; i<=n; i++) {
          lua_pushvalue(L, 1);      /* push function */
          lua_rawgeti(L, 2, i);     /* get element */
          for (j=3; j<=nargs; j++)
            lua_pushvalue(L, j);
          lua_call(L, nargs-1, 1);  /* call function with nargs-1 argument and one result */
          lua_sinsert(L, -2);
        }
      }
      else {  /* dictionary */
        agn_createset(L, agn_ssize(L, 2));
        lua_pushnil(L);
        while (agn_fnext(L, 2, 1)) {
          for (j=3; j<=nargs; j++)
            lua_pushvalue(L, j);
          lua_call(L, nargs-1, 1);  /* call function with nargs-1 argument and one result */
          lua_rawset2(L, -4);       /* store result in a table */
          agn_poptop(L);
        }
      }
      break;
    }
    case LUA_TSEQ: {
      n = agn_seqsize(L, 2);
      agn_createset(L, n);
      for (i=1; i<=n; i++) {
        lua_pushvalue(L, 1);      /* push function */
        lua_seqgeti(L, 2, i);     /* get element */
        for (j=3; j<=nargs; j++)
          lua_pushvalue(L, j);
        lua_call(L, nargs-1, 1);  /* call function with nargs-1 argument and one result */
        lua_sinsert(L, -2);
      }
      break;
    }
    default:
      luaL_error(L, "in " LUA_QS ", table or sequence expected as second argument.", "maptoset");
  }
  return 1;
}



static int agnB_map (lua_State *L) {
  int i, j, n, nargs, nops;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  nargs = lua_gettop(L);
  luaL_checkstack(L, nargs, "too many arguments");
  switch (lua_type(L, 2)) {
    case LUA_TTABLE: {
      n = luaL_getn(L, 2);
      if (n != 0) {  /* simple table */
        lua_createtable(L, n, 0);
        for (i=1; i<=n; i++) {
          lua_pushvalue(L, 1);      /* push function */
          lua_rawgeti(L, 2, i);     /* get element */
          for (j=3; j<=nargs; j++)
            lua_pushvalue(L, j);
          lua_call(L, nargs-1, 1);  /* call function with nargs-1 argument and one result */
          lua_rawseti(L, -2, i);    /* store result in a table */
        }
      } else {  /* dictionary */
        lua_createtable(L, 0, agn_size(L, 2));
        lua_pushnil(L);
        while (agn_fnext(L, 2, 1)) {
          for (j=3; j<=nargs; j++)
            lua_pushvalue(L, j);
          lua_call(L, nargs-1, 1);  /* call function with nargs-1 argument and one result */
          lua_rawset2(L, -4);       /* store result in a table */
          agn_poptop(L);
        }
      }
      /* set type of new table to the one of the original table */
      if (agn_getutype(L, 2)) {
        agn_setutype(L, -2, -1);
        agn_poptop(L);  /* pop type (string) */
      }
      /* copy metatable */
      if (lua_getmetatable(L, 2))
        lua_setmetatable(L, -2);
      break;
    }  /* end of case LUA_TTABLE */
    case LUA_TSET: {
      agn_createset(L, agn_ssize(L, 2));
      lua_pushnil(L);
      while (lua_usnext(L, 2)) {
        lua_pushvalue(L, 1);      /* push function; FIXME: can be optimized */
        lua_pushvalue(L, -2);     /* push key (value) */
        for (j=3; j<=nargs; j++)
          lua_pushvalue(L, j);
        lua_call(L, nargs-1, 1);  /* call function with nargs-1 argument and one result */
        lua_sinsert(L, -4);       /* store result to new set */
        agn_poptop(L);
      }
      /* copy metatable */
      if (lua_getmetatable(L, 2))
        lua_setmetatable(L, -2);
      break;
    } /* end of case LUA_TSET */
    case LUA_TSEQ: {  /* 0.11.0 */
      nops = agn_seqsize(L, 2);
      agn_createseq(L, nops);
      /* set its type to the one of the original sequence */
      if (agn_getutype(L, 2)) {
        agn_setutype(L, -2, -1);
        agn_poptop(L);  /* pop type (string) */
      }
      /* copy metatable */
      if (lua_getmetatable(L, 2))
        lua_setmetatable(L, -2);
      /* now traverse array */
      for (i=0; i < nops; i++) {
        lua_pushvalue(L, 1);      /* push function; FIXME: can be optimized */
        lua_seqgeti(L, 2, i+1);
        for (j=3; j<=nargs; j++)
          lua_pushvalue(L, j);
        lua_call(L, nargs-1, 1);  /* call function with nargs-1 argument and one result */
        lua_seqinsert(L, -2);      /* store result to new sequence */
      }
      break;
    } /* end of case LUA_TSEQ */
    default:
      luaL_error(L, "in " LUA_QS ", table, set, or sequence expected as second argument.", "map");
  } /* end of switch */
  return 1;
}


/* 0.9.1; January 11, 2008; fixed June 30, 2008 */
static int luaB_select (lua_State *L) {
  size_t j, nargs;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  nargs = lua_gettop(L);
  luaL_checkstack(L, nargs, "too many arguments");
  switch(lua_type(L, 2)) {
    case LUA_TTABLE: {
      size_t i, n;
      n = luaL_getn(L, 2);
      if (n == 0) {  /* dictionary */
        lua_createtable(L, 0, agn_size(L, 2));
        lua_pushnil(L);
        while (agn_fnext(L, 2, 1)) {
          for (j=3; j<=nargs; j++)
            lua_pushvalue(L, j);
          lua_call(L, nargs-1, 1);  /* call function with nargs-1 argument and one result */
          if (agn_istrue(L, -1)) {
            agn_poptop(L);  /* delete result */
            lua_rawset2(L, -3);  /* store value in a table */
          } else
            agn_poptoptwo(L);  /* delete result and value */
        }
      } else {  /* simple table */
        int c = 0;
        lua_createtable(L, n, 0);
        for (i=1; i<=n; i++) {
          lua_pushvalue(L, 1);        /* push function */
          lua_rawgeti(L, 2, i);       /* get element */
          for (j=3; j<=nargs; j++)
            lua_pushvalue(L, j);
          lua_call(L, nargs-1, 1);    /* call function with nargs-1 argument and one result */
          if (agn_istrue(L, -1)) {
            lua_rawgeti(L, 2, i);     /* get value again */
            lua_rawseti(L, -3, ++c);  /* store value in new table */
          }
          agn_poptop(L);
        }
      }
      /* set type of new table to the one of the original table */
      if (agn_getutype(L, 2)) {
        agn_setutype(L, -2, -1);
        agn_poptop(L);  /* pop type (string) */
      }
      /* copy metatable */
      if (lua_getmetatable(L, 2))
        lua_setmetatable(L, -2);
      break;
    }
    case LUA_TSET: {
      agn_createset(L, agn_ssize(L, 2));
      /* copy metatable */
      if (lua_getmetatable(L, 2))
        lua_setmetatable(L, -2);
      lua_pushnil(L);
      while (lua_usnext(L, 2)) {
        lua_pushvalue(L, 1);      /* push function; FIXME: can be optimized */
        lua_pushvalue(L, -2);     /* push key (value) */
        for (j=3; j<=nargs; j++)
          lua_pushvalue(L, j);
        lua_call(L, nargs-1, 1);  /* call function with nargs-1 argument and one result */
        if (agn_istrue(L, -1)) {
          agn_poptop(L);
          lua_sinsert(L, -3);    /* set key to new set */
        } else
          agn_poptoptwo(L);
      }
      break;
    }
    case LUA_TSEQ: {  /* 0.11.1 */
      int i, nops;
      nops = agn_seqsize(L, 2);
      agn_createseq(L, nops);
      /* set its type to the one of the original array */
      if (agn_getutype(L, 2)) {
        agn_setutype(L, -2, -1);
        agn_poptop(L);  /* pop type (string) */
      }
      /* copy metatable */
      if (lua_getmetatable(L, 2))
        lua_setmetatable(L, -2);
      /* now traverse sequence */
      for (i=0; i < nops; i++) {
        lua_pushvalue(L, 1);      /* push function */
        lua_seqgeti(L, 2, i+1);
        for (j=3; j<=nargs; j++)
          lua_pushvalue(L, j);
        lua_call(L, nargs-1, 1);  /* call function with nargs-1 argument and one result */
        if (agn_istrue(L, -1)) {
          lua_seqgeti(L, 2, i+1);
          lua_seqinsert(L, -3);   /* store value to new sequence */
        }
        agn_poptop(L);
      }
      break;
    } /* end of case LUA_TSEQ */
    default:
      luaL_error(L, "in " LUA_QS ", table, set, or sequence expected as second argument.", "select");
  } /* end of switch */
  /* FIXME: Should resize the table to actual size */
  return 1;
}


/* 0.10.0; March 30, 2008; extended with sequences on June 30, 2008 */
static int tbl_remove (lua_State *L) {
  size_t j, nargs;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  nargs = lua_gettop(L);
  luaL_checkstack(L, nargs, "too many arguments");
  switch (lua_type(L, 2)) {
    case LUA_TTABLE: {
      size_t i, n;
      n = luaL_getn(L, 2);
      if (n == 0) {  /* dictionary */
        lua_createtable(L, 0, agn_size(L, 2));
        lua_pushnil(L);
        while (agn_fnext(L, 2, 1)) {
          for (j=3; j<=nargs; j++)
            lua_pushvalue(L, j);
          lua_call(L, nargs-1, 1);    /* call function with nargs-1 argument and one result */
          if (agn_isfalse(L, -1)) {
            agn_poptop(L);            /* delete result */
            lua_rawset2(L, -3);       /* store value in a table */
          } else
            agn_poptoptwo(L);         /* delete result and value */
        }
      } else {  /* simple table */
        int c = 0;
        lua_createtable(L, n, 0);
        for (i=1; i<=n; i++) {
          lua_pushvalue(L, 1);        /* push function */
          lua_rawgeti(L, 2, i);       /* get element */
          for (j=3; j<=nargs; j++)
            lua_pushvalue(L, j);
          lua_call(L, nargs-1, 1);    /* call function with nargs-1 argument and one result */
          if (agn_isfalse(L, -1)) {
            lua_rawgeti(L, 2, i);     /* get value again */
            lua_rawseti(L, -3, ++c);  /* store value in new table */
          }
          agn_poptop(L);              /* delete result */
        }
      }
      break;
    } /* end of case LUA_TTABLE */
    case LUA_TSET: {
      agn_createset(L, agn_ssize(L, 2));
      lua_pushnil(L);
      while (lua_usnext(L, 2)) {
        lua_pushvalue(L, 1);      /* push function */
        lua_pushvalue(L, -2);     /* push key (value) */
        for (j=3; j<=nargs; j++)
          lua_pushvalue(L, j);
        lua_call(L, nargs-1, 1);  /* call function with nargs-1 argument and one result */
        if (agn_isfalse(L, -1)) {
          agn_poptop(L);
          lua_sinsert(L, -3);     /* set key to new set */
        } else
          agn_poptoptwo(L);
      }
      break;
    } /* end of case LUA_TSET */
    case LUA_TSEQ: {  /* 0.12.0 */
      int i, nops;
      nops = agn_seqsize(L, 2);
      agn_createseq(L, nops);
      /* set its type to the one of the original array */
      if (agn_getutype(L, 2)) {
        agn_setutype(L, -2, -1);
        agn_poptop(L);  /* pop type (string) */
      }
      /* copy metatable */
      if (lua_getmetatable(L, 2))
        lua_setmetatable(L, -2);
      /* now traverse sequence */
      for (i=0; i < nops; i++) {
        lua_pushvalue(L, 1);      /* push function */
        lua_seqgeti(L, 2, i+1);
        for (j=3; j<=nargs; j++)
          lua_pushvalue(L, j);
        lua_call(L, nargs-1, 1);  /* call function with nargs-1 argument and one result */
        if (agn_isfalse(L, -1)) {
          lua_seqgeti(L, 2, i+1);
          lua_seqinsert(L, -3);   /* store value to new sequence */
        }
        agn_poptop(L);            /* delete result */
      }
      break;
    } /* end of case LUA_TSEQ */
    default:
      luaL_error(L, "in " LUA_QS ", expected a table, set, or sequence as second argument.", "remove");
  }
  /* FIXME: Should resize the table to actual size */
  return 1;
}


/* convert a string to a table, needed by phonetiqs.config;
   extended to general function converting structures to tables 0.12.2, November 10, 2008 */

static int agnB_toTable (lua_State *L) {
  size_t i, l;
  switch (lua_type(L, 1)) {
    case LUA_TSTRING: {
      /* const char *str = luaL_checklstring(L, 1, &l); */
      const char *str = lua_tolstring(L, 1, &l);
      lua_createtable(L, l, 0);
      for (i=1; i<l+1; i++) {
        lua_rawsetilstring(L, -1, i, str++, 1);
      }
      break;
    }
    case LUA_TSET: {
      lua_createtable(L, agn_ssize(L, 1), 0);
      i = 0;
      lua_pushnil(L);
      while (lua_usnext(L, 1)) {
        i++;
        lua_rawseti(L, -3, i);
      }
      break;
    }
    case LUA_TSEQ: {
      size_t nops;
      nops = agn_seqsize(L, 1)+1;
      lua_createtable(L, nops, 0);
      for (i=1; i < nops; i++) {
        lua_seqgeti(L, 1, i);
        lua_rawseti(L, -2, i);
      }
      break;
    }
    default: luaL_error(L, "in " LUA_QS ", expected a table, sequence, or string.", "toTable");
  }
  return 1;
}


/* convert a string to a sequence */

static int agnB_toSeq (lua_State *L) {
  size_t i, l;
  switch (lua_type(L, 1)) {
    case LUA_TSTRING: {
      const char *str = lua_tolstring(L, 1, &l);
      agn_createseq(L, l);
      for (i=1; i<l+1; i++) {
        lua_seqrawsetilstring(L, -1, i, str++, 1);
      }
      break;
    }
    case LUA_TTABLE: {
      i = 0;
      agn_createseq(L, agn_size(L, 1));
      lua_pushnil(L);
      while (lua_next(L, 1) != 0) {
        lua_seqseti(L, -3, ++i);
      }
      break;
    }
    case LUA_TSET: {
      i = 0;
      agn_createseq(L, agn_ssize(L, 1));
      lua_pushnil(L);
      while (lua_usnext(L, 1)) {
        lua_seqseti(L, -3, ++i);
      }
      break;
    }
    default: lua_pushfail(L);
  }
  return 1;
}


/* initializes an rtable with an empty table */
static int agnB_rinit (lua_State *L) {
  luaL_checktype(L, 1, LUA_TFUNCTION);
  agn_creatertable(L, 1);
  lua_pushnil(L);
  return 1;
}


/* returns the remember table */
static int agnB_rget (lua_State *L) {
  luaL_checktype(L, 1, LUA_TFUNCTION);
  luaL_checkany(L, 1);
  if (!agn_getrtable(L, 1)) {
    lua_pushfail(L);  /* no remtable */
  } else {
    lua_newtable(L);
    lua_pushnil(L);
    while (lua_next(L, -3) != 0) {  /* traverse rtable */
      lua_pushnil(L);
      while (lua_next(L, -2) != 0) {  /* get arguments and results */
        lua_rawset2(L, -5);
      }
      agn_poptop(L);
    }
  }
  return 1;
}


/* sets a remember table with initial values */
static int agnB_rset (lua_State *L) {
  agn_setrtable(L, 1, 2, 3);
  lua_pushnil(L);
  return 1;
}


/* deletes an rtable */
static int agnB_rdelete (lua_State *L) {
  luaL_checktype(L, 1, LUA_TFUNCTION);
  agn_deletertable(L, 1);
  lua_pushnil(L);
  return 1;
}


static int agnB_hasrtable (lua_State *L) {
  luaL_checktype(L, 1, LUA_TFUNCTION);
  if (!agn_getrtable(L, 1))
    lua_pushboolean(L, 0);
  else {
    agn_poptop(L);
    lua_pushboolean(L, 1);
  }
  return 1;
}


static int luaB_ops (lua_State *L) {
  int n = lua_gettop(L);
  if (lua_type(L, 1) == LUA_TSTRING && *lua_tostring(L, 1) == '#') {
    lua_pushinteger(L, n-1);
    return 1;
  }
  else {
    int i = luaL_checkint(L, 1);
    if (i < 0) i = n + i;
    else if (i > n) i = n;
    luaL_argcheck(L, 1 <= i, 1, "index out of range");
    return n - i;
  }
}


static int luaB_attrib (lua_State *L) {  /* added 0.10.0 */
  switch (lua_type(L, 1)) {
    case LUA_TTABLE: {
      size_t a[4];
      lua_newtable(L);
      agn_tablestate(L, 1, a);
      lua_pushstring(L, "array_assigned");
      lua_pushnumber(L, a[0]);
      lua_rawset(L, -3);
      lua_pushstring(L, "hash_assigned");
      lua_pushnumber(L, a[1]);
      lua_rawset(L, -3);
      lua_pushstring(L, "array_allocated");
      lua_pushnumber(L, a[2]);
      lua_rawset(L, -3);
      lua_pushstring(L, "hash_allocated");
      lua_pushnumber(L, a[3]);
      lua_rawset(L, -3);
      break;
    }
    case LUA_TSET: {
      size_t a[2];
      lua_newtable(L);
      agn_sstate(L, 1, a);
      lua_pushstring(L, "hash_assigned");
      lua_pushnumber(L, a[0]);
      lua_rawset(L, -3);
      lua_pushstring(L, "hash_allocated");
      lua_pushnumber(L, a[1]);
      lua_rawset(L, -3);
      break;
    }
    case LUA_TSEQ: {
      size_t a[2];
      lua_newtable(L);
      agn_seqstate(L, 1, a);
      lua_pushstring(L, "size");
      lua_pushnumber(L, a[0]);
      lua_rawset(L, -3);
      lua_pushstring(L, "maxsize");
      lua_pushnumber(L, a[1]);
      lua_rawset(L, -3);
      break;
    }
    default:
      luaL_error(L, "in " LUA_QS ", table, set, or sequence expected.", "attrib");
  }
  return 1;
}


static int luaB_settype (lua_State *L) {  /* added 0.11.0, extended 0.12.0, 0.12.2, 0.14.0  */
  int i, s, t, nargs;
  nargs = lua_gettop(L);
  luaL_checkstack(L, nargs, "too many arguments");
  if (nargs < 2)
    luaL_error(L, "in " LUA_QS ", at least two arguments expected.", "settype");
  t = lua_type(L, nargs);
  luaL_argcheck(L, t == LUA_TSTRING || t == LUA_TNIL, nargs, "expected a string or null");
  for (i=1; i<nargs; i++) {
    s = lua_type(L, i);
    luaL_argcheck(L, s == LUA_TSEQ || s == LUA_TPAIR || s == LUA_TTABLE || s == LUA_TSET || s == LUA_TFUNCTION,
      i, "expected a procedure, sequence, set, pair, or table");
    agn_setutype(L, i, nargs);
  }
  lua_pushnil(L);
  return 1;
}


static int luaB_gettype (lua_State *L) {  /* added 0.11.0, changed 0.12.2; extended 0.14.0 */
  int s = lua_type(L, 1);
  if (!(s == LUA_TSEQ || s == LUA_TPAIR || s == LUA_TTABLE || s == LUA_TFUNCTION || s == LUA_TSET) || !agn_getutype(L, 1))
    lua_pushnil(L);
  return 1;
}

/* added to baselib in 0.11.0; modified 0.9.2: tuned by 45 % */

static int luaB_concat (lua_State *L) {
  luaL_Buffer b;
  size_t l, lsep;
  int i, last, type;
  const char *str, *sep;  /* added 0.9.2 */
  sep = luaL_optlstring(L, 2, "", &lsep);
  i = luaL_optint(L, 3, 1);
  type = lua_type(L, 1);
  if (type == LUA_TTABLE) {
    last = luaL_opt(L, luaL_checkint, 4, luaL_getn(L, 1));
    luaL_buffinit(L, &b);
    for (; i <= last; i++) {
      str = agn_getistring(L, 1, i);
      l = strlen(str);
      if (str == NULL) continue;  /* Agena crashes otherwise */
      luaL_addlstring(&b, str, l);
      if (i != last)
        luaL_addlstring(&b, sep, lsep);
    }
    luaL_pushresult(&b);
  }
  else if (type == LUA_TSEQ) {
    last = luaL_opt(L, luaL_checkint, 4, agn_seqsize(L, 1));
    luaL_buffinit(L, &b);
    for (; i <= last; i++) {
      str = agn_getseqlstring(L, 1, i, &l);
      if (str == NULL) continue;  /* Agena crashes otherwise */
      luaL_addlstring(&b, str, l);
      if (i != last)
        luaL_addlstring(&b, sep, lsep);
    }
    luaL_pushresult(&b);
  }
  else
    luaL_error(L, "in " LUA_QS ", table or sequence expected.", "concat");
  return 1;
}


/*
** {======================================================
** Quicksort
** (based on `Algorithms in MODULA-3', Robert Sedgewick;
**  Addison-Wesley, 1993.)
*/

#define luaBset2(L, i, j) { \
  lua_rawseti(L, 1, (i)); \
  lua_rawseti(L, 1, (j)); \
}


#define luaBsetseq2(L, i, j) { \
  lua_seqseti(L, 1, (i)); \
  lua_seqseti(L, 1, (j)); \
}


int luaBsort_comp (lua_State *L, int a, int b) {
  if (!lua_isnil(L, 2)) {  /* function? */
    int res;
    lua_pushvalue(L, 2);
    lua_pushvalue(L, a-1);  /* -1 to compensate function */
    lua_pushvalue(L, b-2);  /* -2 to compensate function and `a' */
    lua_call(L, 2, 1);
    res = lua_toboolean(L, -1);
    lua_pop(L, 1);
    return res;
  }
  else  /* a < b? */
    return lua_lessthan(L, a, b);
}


static void auxsorttable (lua_State *L, int l, int u) {
  while (l < u) {  /* for tail recursion */
    int i, j;
    /* sort elements a[l], a[(l+u)/2] and a[u] */
    lua_rawgeti(L, 1, l);
    lua_rawgeti(L, 1, u);
    if (luaBsort_comp(L, -1, -2)) { /* a[u] < a[l]? */
      luaBset2(L, l, u);  /* swap a[l] - a[u] */
    } else
      lua_pop(L, 2);
    if (u-l == 1) break;  /* only 2 elements */
    i = (l+u)/2;
    lua_rawgeti(L, 1, i);
    lua_rawgeti(L, 1, l);
    if (luaBsort_comp(L, -2, -1)) { /* a[i]<a[l]? */
      luaBset2(L, i, l);
    } else {
      lua_pop(L, 1);  /* remove a[l] */
      lua_rawgeti(L, 1, u);
      if (luaBsort_comp(L, -1, -2)) { /* a[u]<a[i]? */
        luaBset2(L, i, u);
      } else
        lua_pop(L, 2);
    }
    if (u-l == 2) break;  /* only 3 elements */
    lua_rawgeti(L, 1, i);  /* Pivot */
    lua_pushvalue(L, -1);
    lua_rawgeti(L, 1, u-1);
    luaBset2(L, i, u-1);
    /* a[l] <= P == a[u-1] <= a[u], only need to sort from l+1 to u-2 */
    i = l; j = u-1;
    for (;;) {  /* invariant: a[l..i] <= P <= a[j..u] */
      /* repeat ++i until a[i] >= P */
      while (lua_rawgeti(L, 1, ++i), luaBsort_comp(L, -1, -2)) {
        if (i>u) luaL_error(L, "invalid order function for sorting");
        lua_pop(L, 1);  /* remove a[i] */
      }
      /* repeat --j until a[j] <= P */
      while (lua_rawgeti(L, 1, --j), luaBsort_comp(L, -3, -1)) {
        if (j<l) luaL_error(L, "invalid order function for sorting");
        lua_pop(L, 1);  /* remove a[j] */
      }
      if (j<i) {
        lua_pop(L, 3);  /* pop pivot, a[i], a[j] */
        break;
      }
      luaBset2(L, i, j);
    }
    lua_rawgeti(L, 1, u-1);
    lua_rawgeti(L, 1, i);
    luaBset2(L, u-1, i);  /* swap pivot (a[u-1]) with a[i] */
    /* a[l..i-1] <= a[i] == P <= a[i+1..u] */
    /* adjust so that smaller half is in [j..i] and larger one in [l..u] */
    if (i-l < u-i) {
      j=l; i=i-1; l=i+2;
    }
    else {
      j=i+1; i=u; u=j-2;
    }
    auxsorttable(L, j, i);  /* call recursively the smaller one */
  }  /* repeat the routine for the larger one */
}


static void auxsortseq (lua_State *L, int l, int u) {
  while (l < u) {  /* for tail recursion */
    int i, j;
    /* sort elements a[l], a[(l+u)/2] and a[u] */
    lua_seqgeti(L, 1, l);
    lua_seqgeti(L, 1, u);
    if (luaBsort_comp(L, -1, -2)) { /* a[u] < a[l]? */
      luaBsetseq2(L, l, u);  /* swap a[l] - a[u] */
    } else
      lua_pop(L, 2);
    if (u-l == 1) break;  /* only 2 elements */
    i = (l+u)/2;
    lua_seqgeti(L, 1, i);
    lua_seqgeti(L, 1, l);
    if (luaBsort_comp(L, -2, -1)) { /* a[i]<a[l]? */
      luaBsetseq2(L, i, l);
    } else {
      lua_pop(L, 1);  /* remove a[l] */
      lua_seqgeti(L, 1, u);
      if (luaBsort_comp(L, -1, -2)) { /* a[u]<a[i]? */
        luaBsetseq2(L, i, u);
      } else
        lua_pop(L, 2);
    }
    if (u-l == 2) break;  /* only 3 elements */
    lua_seqgeti(L, 1, i);  /* Pivot */
    lua_pushvalue(L, -1);
    lua_seqgeti(L, 1, u-1);
    luaBsetseq2(L, i, u-1);
    /* a[l] <= P == a[u-1] <= a[u], only need to sort from l+1 to u-2 */
    i = l; j = u-1;
    for (;;) {  /* invariant: a[l..i] <= P <= a[j..u] */
      /* repeat ++i until a[i] >= P */
      while (lua_seqgeti(L, 1, ++i), luaBsort_comp(L, -1, -2)) {
        if (i>u) luaL_error(L, "invalid order function for sorting");
        lua_pop(L, 1);  /* remove a[i] */
      }
      /* repeat --j until a[j] <= P */
      while (lua_seqgeti(L, 1, --j), luaBsort_comp(L, -3, -1)) {
        if (j<l) luaL_error(L, "invalid order function for sorting");
        lua_pop(L, 1);  /* remove a[j] */
      }
      if (j<i) {
        lua_pop(L, 3);  /* pop pivot, a[i], a[j] */
        break;
      }
      luaBsetseq2(L, i, j);
    }
    lua_seqgeti(L, 1, u-1);
    lua_seqgeti(L, 1, i);
    luaBsetseq2(L, u-1, i);  /* swap pivot (a[u-1]) with a[i] */
    /* a[l..i-1] <= a[i] == P <= a[i+1..u] */
    /* adjust so that smaller half is in [j..i] and larger one in [l..u] */
    if (i-l < u-i) {
      j=l; i=i-1; l=i+2;
    }
    else {
      j=i+1; i=u; u=j-2;
    }
    auxsortseq(L, j, i);  /* call recursively the smaller one */
  }  /* repeat the routine for the larger one */
}


static int luaB_sort (lua_State *L) {
  int n = agn_nops(L, 1);
  luaL_checkstack(L, 40, "");  /* assume array is smaller than 2^40 */
  if (!lua_isnoneornil(L, 2))  /* is there a 2nd argument? */
    luaL_checktype(L, 2, LUA_TFUNCTION);
  lua_settop(L, 2);  /* make sure there is two arguments */
  switch (lua_type(L, 1)) {
    case LUA_TTABLE:
      auxsorttable(L, 1, n);
      break;
    case LUA_TSEQ:
      auxsortseq(L, 1, n);
      break;
    default:
      luaL_error(L, "in " LUA_QS ", table or sequence expected as first argument.", "sort");
  }
  return 0;
}


static int agnB_zip (lua_State *L) {
  int i, sizea, sizeb;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  if (lua_type(L, 2) == LUA_TSEQ && lua_type(L, 3) == LUA_TSEQ) {
    sizea = agn_seqsize(L, 2);
    sizeb = agn_seqsize(L, 3);
    if (sizea != sizeb)
      luaL_error(L, "in " LUA_QS ", sequences of different size.", "zip");
    agn_createseq(L, sizea);
    /* set its type to the one of the original sequence */
    if (agn_getutype(L, 2)) {
      agn_setutype(L, -2, -1);
      agn_poptop(L);  /* pop type (string) */
    } else if (agn_getutype(L, 3)) {
      agn_setutype(L, -2, -1);
      agn_poptop(L);  /* pop type (string) */
    }
    /* copy metatable */
    if (lua_getmetatable(L, 2))
      lua_setmetatable(L, -2);
    else if (lua_getmetatable(L, 3))
      lua_setmetatable(L, -2);
    /* now traverse array */
    for (i=0; i < sizea; i++) {
      lua_pushvalue(L, 1);      /* push function; FIXME: can be optimized */
      lua_seqgeti(L, 2, i+1);
      lua_seqgeti(L, 3, i+1);
      lua_call(L, 2, 1);        /* call function with 2 arguments and one result */
      lua_seqinsert(L, -2);     /* store result to new sequence */
    }
  }
  else if (lua_type(L, 2) == LUA_TTABLE && lua_type(L, 3) == LUA_TTABLE) {
    sizea = agn_size(L, 2);
    sizeb = agn_size(L, 3);
    if (sizea != sizeb)
      luaL_error(L, "in " LUA_QS ", tables of different size.", "zip");
    lua_createtable(L, sizea, 0);
    /* set its type to the one of the original table */
    if (agn_getutype(L, 2)) {
      agn_setutype(L, -2, -1);
      agn_poptop(L);  /* pop type (string) */
    } else if (agn_getutype(L, 3)) {
      agn_setutype(L, -2, -1);
      agn_poptop(L);  /* pop type (string) */
    }
    /* copy metatable */
    if (lua_getmetatable(L, 2))
      lua_setmetatable(L, -2);
    else if (lua_getmetatable(L, 3))
      lua_setmetatable(L, -2);
    /* now traverse tables */
    for (i=0; i < sizea; i++) {
      lua_pushvalue(L, 1);      /* push function; FIXME: can be optimized */
      lua_rawgeti(L, 2, i+1);
      lua_rawgeti(L, 3, i+1);
      lua_call(L, 2, 1);        /* call function with 2 arguments and one result */
      lua_rawseti(L, -2, i+1);  /* store result to new table */
    }
  } else
    luaL_error(L, "in " LUA_QS ", expected either two sequences or two tables.", "zip");
  return 1;
}


/* luaB_min: return the smallest value in a list. If the option 'sorted' is given,
   the list is not traversed, instead the first entry in the list is returned.
   If the list is empty, NULL is returned. lua_getn must be called at first,
   otherwise lua_rawgeti would return 0 with empty lists.
   June 20, 2007 */

static int luaB_min (lua_State *L) {
  int i, n, type;
  lua_Number min;
  const char *option;
  type = lua_type(L, 1);
  luaL_argcheck(L, type == LUA_TTABLE || type == LUA_TSEQ, 1, "table or sequence expected");
  option = luaL_optstring(L, 2, "default");
  n = agn_nops(L, 1);
  min = 0;  /* to prevent compiler warnings */
  if (strcmp(option, "sorted") == 0) {
    if (n == 0) {
      lua_pushnil(L);
      return 1;
    }
    lua_pushnumber(L, (type == LUA_TTABLE) ? agn_getinumber(L, 1, 1) : lua_seqgetinumber(L, 1, 1) );
    return 1;
  }
  switch (type) {
    case LUA_TTABLE: {
      min = agn_getinumber(L, 1, 1);
      for (i=2; i<=n; i++) {
        lua_Number v = agn_getinumber(L, 1, i);
        if (v < min)
          min = v;
      }
      break;
    }
    case LUA_TSEQ: {
      min = lua_seqgetinumber(L, 1, 1);
      for (i=2; i<=n; i++) {
        lua_Number v = lua_seqgetinumber(L, 1, i);
        if (v < min)
          min = v;
      }
      break;
    }
    default: lua_assert(0);
  }
  lua_pushnumber(L, min);
  return 1;
}


/* tbl_max: return the largest value in a list. If the option 'sorted' is given,
   the list is not traversed, instead the last entry in the list is returned.
   If the list is empty, NULL is returned. lua_getn must be called at first,
   otherwise lua_rawgeti would return 0 with empty lists.
   June 20, 2007; tuned December 23, 2007 */

static int luaB_max (lua_State *L) {
  int i, n, type;
  lua_Number max;
  const char *option;
  type = lua_type(L, 1);
  luaL_argcheck(L, type == LUA_TTABLE || type == LUA_TSEQ, 1, "table or sequence expected");
  option = luaL_optstring(L, 2, "default");
  n = agn_nops(L, 1);
  max = 0;  /* to prevent compiler warnings */
  if (strcmp(option, "sorted") == 0) {
    if (n == 0) {
      lua_pushnil(L);
      return 1;
    }
    lua_pushnumber(L, (type == LUA_TTABLE) ? agn_getinumber(L, 1, n) : lua_seqgetinumber(L, 1, n) );
    return 1;
  }
  switch (type) {
    case LUA_TTABLE: {
      max = agn_getinumber(L, 1, 1);
      for (i=2; i<=n; i++) {
        lua_Number v = agn_getinumber(L, 1, i);
        if (v > max)
          max = v;
      }
      break;
    }
    case LUA_TSEQ: {
      max = lua_seqgetinumber(L, 1, 1);
      for (i=2; i<=n; i++) {
        lua_Number v = lua_seqgetinumber(L, 1, i);
        if (v > max)
          max = v;
      }
      break;
    }
    default: lua_assert(0);
  }
  lua_pushnumber(L, max);
  return 1;
}


/*static int luaB_test (lua_State *L) {
  int e;
  lua_pushnumber(L, agn_tonumber(L, 1));
//  lua_pushnumber(L, agn_tonumberx(L, 1, &e));
  return 2;
}*/


static const luaL_Reg base_funcs[] = {
  {"assume", luaB_assume},
  {"gc", luaB_collectgarbage},  /* formerly {"collectgarbage", ...},  changed 0.4.0 */
  {"run", luaB_run},
  {"error", luaB_error},
  {"gcinfo", luaB_gcinfo},
  {"getfenv", luaB_getfenv},
  {"getmetatable", luaB_getmetatable},
  {"loadfile", luaB_loadfile},
  {"load", luaB_load},
  {"loadstring", luaB_loadstring},
  {"next", luaB_next},
  {"pcall", luaB_pcall},
  {"print", luaB_print},
  {"__DOTTYPRINT", luaB_dottyprint},  /* added July 01, 2008 */
  {"mprint", luaB_mprint},
  {"rawequal", luaB_rawequal},
  {"rawget", luaB_rawget},
  {"rawset", luaB_rawset},
  {"select", luaB_select},
  {"remove", tbl_remove},        /* added March 30, 2008 */
  {"setfenv", luaB_setfenv},
  {"setmetatable", luaB_setmetatable},
  {"toNumber", luaB_toNumber},
  {"toString", luaB_toString},
  {"toTable", agnB_toTable},     /* added June 08, 2007 */
  {"toSeq", agnB_toSeq},         /* added June 14, 2008 */
  {"unpack", luaB_unpack},
  {"xpcall", luaB_xpcall},
  {"__RESTART", agn_restart},    /* added December 16, 2006 */
  {"time", agn_time},            /* added */
  {"readlib", agn_readlib},      /* added December 10, 2006 */
  {"rinit", agnB_rinit},         /* added January 28, 2008 */
  {"rset", agnB_rset},           /* added January 28, 2008 */
  {"rget", agnB_rget},           /* added January 28, 2008 */
  {"rdelete", agnB_rdelete},     /* added February 02, 2008 */
  {"hasrtable", agnB_hasrtable}, /* added February 05, 2008 */
  {"ops", luaB_ops},
  {"map", agnB_map},             /* added April 07, 2007 */
  {"maptoset", agnB_maptoset},   /* added April 11, 2008 */
  {"attrib", luaB_attrib},       /* added April 20, 2008 */
  {"settype", luaB_settype},     /* added May 05, 2008 */
  {"gettype", luaB_gettype},     /* added May 05, 2008 */
  {"concat", luaB_concat},       /* added May 17, 2008 */
  {"sort", luaB_sort},
  {"loadClib", agn_loadClib},    /* added July 08, 2008 */
  {"zip", agnB_zip},             /* added September 04, 2008 */
  {"min", luaB_min},
  {"max", luaB_max},
  /* {"test", luaB_test}, */
  {NULL, NULL}
};


/*
** {======================================================
** Coroutine library
** =======================================================
*/

static int auxresume (lua_State *L, lua_State *co, int narg) {
  int status;
  if (!lua_checkstack(co, narg))
    luaL_error(L, "too many arguments to resume");
  if (lua_status(co) == 0 && lua_gettop(co) == 0) {
    lua_pushliteral(L, "cannot resume dead coroutine");
    return -1;  /* error flag */
  }
  lua_xmove(L, co, narg);
  status = lua_resume(co, narg);
  if (status == 0 || status == LUA_YIELD) {
    int nres = lua_gettop(co);
    /* if (!lua_checkstack(L, nres)) 5.1.4 patch 2 */
    if (!lua_checkstack(L, nres + 1))
      luaL_error(L, "too many results to resume");
    lua_xmove(co, L, nres);  /* move yielded values */
    return nres;
  }
  else {
    lua_xmove(co, L, 1);  /* move error message */
    return -1;  /* error flag */
  }
}


static int luaB_coresume (lua_State *L) {
  lua_State *co = lua_tothread(L, 1);
  int r;
  luaL_argcheck(L, co, 1, "coroutine expected");
  r = auxresume(L, co, lua_gettop(L) - 1);
  if (r < 0) {
    lua_pushfalse(L);
    lua_insert(L, -2);
    return 2;  /* return false + error message */
  }
  else {
    lua_pushtrue(L);
    lua_insert(L, -(r + 1));
    return r + 1;  /* return true + `resume' returns */
  }
}


static int luaB_auxwrap (lua_State *L) {
  lua_State *co = lua_tothread(L, lua_upvalueindex(1));
  int r = auxresume(L, co, lua_gettop(L));
  if (r < 0) {
    if (lua_isstring(L, -1)) {  /* error object is a string? */
      luaL_where(L, 1);  /* add extra info */
      lua_insert(L, -2);
      lua_concat(L, 2);
    }
    lua_error(L);  /* propagate error */
  }
  return r;
}


static int luaB_cocreate (lua_State *L) {
  lua_State *NL = lua_newthread(L);
  luaL_argcheck(L, lua_isfunction(L, 1) && !lua_iscfunction(L, 1), 1,
    "Agena function expected");
  lua_pushvalue(L, 1);  /* move function to top */
  lua_xmove(L, NL, 1);  /* move function from L to NL */
  return 1;
}


static int luaB_cowrap (lua_State *L) {
  luaB_cocreate(L);
  lua_pushcclosure(L, luaB_auxwrap, 1);
  return 1;
}


static int luaB_yield (lua_State *L) {
  return lua_yield(L, lua_gettop(L));
}


static int luaB_costatus (lua_State *L) {
  lua_State *co = lua_tothread(L, 1);
  luaL_argcheck(L, co, 1, "coroutine expected");
  if (L == co) lua_pushliteral(L, "running");
  else {
    switch (lua_status(co)) {
      case LUA_YIELD:
        lua_pushliteral(L, "suspended");
        break;
      case 0: {
        lua_Debug ar;
        if (lua_getstack(co, 0, &ar) > 0)  /* does it have frames? */
          lua_pushliteral(L, "normal");  /* it is running */
        else if (lua_gettop(co) == 0)
            lua_pushliteral(L, "dead");
        else
          lua_pushliteral(L, "suspended");  /* initial state */
        break;
      }
      default:  /* some error occured */
        lua_pushliteral(L, "dead");
        break;
    }
  }
  return 1;
}


static int luaB_corunning (lua_State *L) {
  if (lua_pushthread(L))
    return 0;  /* main thread is not a coroutine */
  else
    return 1;
}


static const luaL_Reg co_funcs[] = {
  {"create", luaB_cocreate},
  {"resume", luaB_coresume},
  {"running", luaB_corunning},
  {"status", luaB_costatus},
  {"wrap", luaB_cowrap},
  {"yield", luaB_yield},
  {NULL, NULL}
};

/* }====================================================== */


static void base_open (lua_State *L) {
  /* set global _G */
  lua_pushvalue(L, LUA_GLOBALSINDEX);
  lua_setglobal(L, "_G");
  /* define PI and EXP */
  lua_pushnumber(L, PI);
  lua_setglobal(L, "Pi");
  lua_pushnumber(L, _E);
  lua_setglobal(L, "Exp");
  lua_pushnumber(L, AGN_EPSILON);
  lua_setglobal(L, "_EnvEps");
  lua_pushnumber(L, HUGE_VAL);
  lua_setglobal(L, "infinity");
  lua_pushnumber(L, AGN_NAN);
  lua_setglobal(L, "undefined");
  lua_pushnumber(L, RADIANS_PER_DEGREE);
  lua_setglobal(L, "radians");
  lua_pushnumber(L, 1/RADIANS_PER_DEGREE);
  lua_setglobal(L, "degrees");
  lua_pushnumber(L, LONG_MAX);
  lua_setglobal(L, "_EnvMaxLong");
  lua_pushnumber(L, LONG_MIN);
  lua_setglobal(L, "_EnvMinLong");
  /* open lib into global table */
  luaL_register(L, "_G", base_funcs);
  lua_pushliteral(L, AGENA_RELEASE);                              /* added */
  lua_setglobal(L, "_RELEASE");  /* set global _RELEASE */
  lua_pushliteral(L, AGENA_VERSION);                              /* added 0.14.0 */
  lua_setglobal(L, "_VERSION");  /* set global _VERSION */
  lua_createtable(L, 0, 0);
  lua_setglobal(L, "_EnvPrint");
  /* `newproxy' needs a weaktable as upvalue */
  lua_createtable(L, 0, 1);  /* new table `w' */
  lua_pushvalue(L, -1);  /* `w' will be its own metatable */
  lua_setmetatable(L, -2);
  lua_pushliteral(L, "kv");
  lua_setfield(L, -2, "__mode");  /* metatable(w).__mode = "kv" */
  lua_pushcclosure(L, luaB_newproxy, 1);
  lua_setglobal(L, "newproxy");  /* set global `newproxy' */
  /* set protected names for `with` function */
  agn_createset(L, 2);
  lua_sinsertstring(L, -1, "next");
  lua_sinsertstring(L, -1, "print");
  lua_setglobal(L, "_EnvWithProtected");
  /* set number of lines to be printed at once at table output */
  lua_pushinteger(L, 40);
  lua_setglobal(L, "_EnvMore");
}


LUALIB_API int luaopen_base (lua_State *L) {
  base_open(L);
  luaL_register(L, LUA_COLIBNAME, co_funcs);
  return 2;
}


